/*
** File: model_admxrc6tl.c  
** Project: ADB3 core driver
** Purpose: ADM-XRC-6TL-specific code, including function to bootstrap device context.
**
** (C) Copyright Alpha Data 2010
**
** NOTES
**
** 1. Organization of Flash device U32 (all offsets in bytes):
**
**  0x0000000 - 0x07FFFFF  Alternate ADB3 PCI-E to OCP Bridge bitstream
**  0x0800000 - 0x0FFFFFF  Default ADB3 PCI-E to OCP Bridge bitstream
**  0x1000000 - 0x10FFFFF  Alpha Data VPD region
**  0x1100000 - 0x11FFFFF  Customer VPD region
**  0x1200000 - 0x28FFFFF  Target FPGA bitstream
**  0x2900000 - 0x3FFFFFF  Failsafe target FPGA bitstream
*/

#include <df.h>

#include "adb3.h"
#include "adb3_common.h"
#include "admxrc6tl.h"
#include "admxrc6tx_common.h"
#include "coredma.h"
#include "device.h"
#include "model_boot.h"
#include "flash.h"
#include "flash_cfi.h"

static boolean_t
readVpd(
	Adb3CoreDeviceContext* pDevCtx)
{
  VpdAdmxrc6tlRev1* pVpd = &pDevCtx->info.vpd.data.admxrc6tl.rev1;
  union {
    VpdAdmxrc6tlRev0 rev0;
    VpdAdmxrc6tlRev1 rev1;
  } vpdRaw;
  uint16_t vpdVersion = 0xffffU;
  uint16_t vpdLength = 0U;
  int i;

  flashRead(pDevCtx, 0, pDevCtx->info.bootstrap.vpdOffset, sizeof(vpdVersion), &vpdVersion);
  vpdVersion = dfLe16ToCpu(vpdVersion);
  flashRead(pDevCtx, 0, pDevCtx->info.bootstrap.vpdOffset + 2, sizeof(vpdLength), &vpdLength);
  vpdLength = dfLe16ToCpu(vpdLength);
  switch (vpdVersion) {
  case 0:
    flashRead(pDevCtx, 0, pDevCtx->info.bootstrap.vpdOffset, sizeof(vpdRaw.rev0), &vpdRaw.rev0);
    break;

  case 1:
    flashRead(pDevCtx, 0, pDevCtx->info.bootstrap.vpdOffset, sizeof(vpdRaw.rev1), &vpdRaw.rev1);
    break;

  case 0xffffU:
  	dfDebugPrint(0, ("*** readVpd: invalid VPD version %lu(0x%lx)\n",
      (unsigned long)vpdVersion, (unsigned long)vpdVersion));
    return FALSE;

  default:
  	dfDebugPrint(0, ("+++ readVpd: unrecognised VPD version %lu(0x%lx)\n",
      (unsigned long)vpdVersion, (unsigned long)vpdVersion));
    /* Assume rev1 */
    flashRead(pDevCtx, 0, pDevCtx->info.bootstrap.vpdOffset, sizeof(vpdRaw.rev1), &vpdRaw.rev1);
    break;
  }

  adb3VerifyChecksum(pDevCtx, vpdLength);

  DF_ZERO_OBJECT(pVpd);
  pVpd->version = vpdVersion;
  pVpd->length = vpdLength;
  pVpd->serial = dfLe32ToCpu(vpdRaw.rev0.serial);
  pVpd->modifications = dfLe32ToCpu(vpdRaw.rev0.modifications);
  pVpd->lclkRef = dfLe32ToCpu(vpdRaw.rev0.lclkRef);
  for (i = 0; i < 4; i++) {
    pVpd->sdram[i].order = dfLe8ToCpu(vpdRaw.rev0.sdram[i].order);
    pVpd->sdram[i].speed = dfLe16ToCpu(vpdRaw.rev0.sdram[i].speed);
    pVpd->sdram[i].link = dfLe16ToCpu(vpdRaw.rev0.sdram[i].link);
  }
  pVpd->fpga.device = dfLe16ToCpu(vpdRaw.rev0.fpga.device);
  dfAssert(sizeof(char) == sizeof(uint8_t));
  dfStrCpy((char*)pVpd->fpga.speedGrade, DF_ARRAY_LENGTH(pVpd->fpga.speedGrade), (const char*)vpdRaw.rev0.fpga.speedGrade);
  dfStrCpy((char*)pVpd->fpga.stepping, DF_ARRAY_LENGTH(pVpd->fpga.stepping), (const char*)vpdRaw.rev0.fpga.stepping);
  pVpd->fpga.tempGrade = dfLe8ToCpu(vpdRaw.rev0.fpga.tempGrade);
  pVpd->pcbRev = dfLe8ToCpu(vpdRaw.rev0.pcbRev);
  pVpd->cpldRev = dfLe8ToCpu(vpdRaw.rev0.cpldRev);
  pVpd->checksum = dfLe8ToCpu(vpdRaw.rev0.checksum);
  if (vpdVersion >= 1) {
    pVpd->mgtRef = dfLe32ToCpu(vpdRaw.rev1.mgtRef);
  } else {
    pVpd->mgtRef = 125000000U;
  }
  pDevCtx->info.vpd.bValid = TRUE;
  pDevCtx->info.vpd.length = sizeof(pDevCtx->info.vpd.data.admxrc6tl.rev1);

  return TRUE;
}


static ClockProgramStatus
clockProgram(
	Adb3CoreDeviceContext*	pDevCtx,
  unsigned int clockIndex,
	uint32_t clockWord,
	boolean_t bAfter)
{
	ModelRegsAdmxrc6tx* pModelRegs;
  DfMemoryHandle hModel;
	uint32_t val32, mask;

	dfDebugPrint(5, ("clockProgram: entered, clockIndex=%lu clockWord=0x%lx bAfter=%s\n",
    (unsigned long)clockIndex, (unsigned long)clockWord, bAfter ? "TRUE" : "FALSE"));
	
  pModelRegs = pDevCtx->hardware.model.pAdmxrc6tx;
  hModel = pDevCtx->hardware.hModel;

	if (bAfter) {
    switch (clockIndex) {
    case 0:
			val32 = dfPciMemRead32(hModel, &pModelRegs->lclkStat3);
  	  dfDebugPrint(5, ("clockProgram: LCLKSTAT3=0x%lx\n", (unsigned long)val32));	
			mask = (val32 & 0x4U) ? 0x2U : 0x1U;
			if (!(val32 & mask)) {
				/* DCM not locked */
				return ClockProgramHardwareError;
			}
			if (val32 & 0x28U) {
				/* Control logic busy or write of DCM parameters timed out */
				return ClockProgramHardwareError;
			}
			if (val32 & 0x10U) {
				/* Error */
				return ClockProgramHardwareError;
			}
      break;

    default:
      return ClockProgramInvalidIndex;
		}
	} else {
    switch (clockIndex) {
    case 0:
			dfPciMemWrite32(hModel, &pModelRegs->lclkCtl, clockWord | (0x1U << 31));
      dfPciMemRead32(hModel, &pModelRegs->lclkStat3);
      break;

    default:
      return ClockProgramInvalidIndex;
    }
	}

	return ClockProgramSuccess;
}

static ClockWordStatus
clockWordLclk(
	Adb3CoreDeviceContext* pDevCtx,
  uint32_t flags,
	uint32_t refFrequency,
	uint64_t frequencyReq,
	uint64_t frequencyMin,
	uint64_t frequencyMax,
	uint32_t* pClockWord,
	uint64_t* pActualFrequency)
{
	const uint32_t minFrequency = 31999000, maxFrequency = 210001000;
	uint32_t m, bestM = 0, d, bestD = 0; /* init. to avoid compiler whine */
	uint32_t f, bestFrequency = 0; /* init. to avoid compiler whine */
	uint32_t error, bestError = (uint32_t)-1;

	dfDebugPrint(5, ("clockWordLclk: refFrequency=%lu reqFrequency=%llu\n",
    (unsigned long)refFrequency, (unsigned long long)frequencyReq));

  if (refFrequency == 0 || refFrequency > 268000000) {
		return ClockWordGeneralFailure;
	}
	if (frequencyReq < minFrequency || frequencyReq > maxFrequency) {
		return ClockWordOutOfRange;
	}

	/* Won't overflow assuming (refFrequency <= 268e6) */
	for (d = 1; d <= 32; d++) {
		for (m = 2; m <= 32; m++) {
			f = ((refFrequency >> 1) * m / d) << 1;
      if ((flags & CoreClockMinimumFrequency) && f < frequencyMin) {
        continue;
      }
      if ((flags & CoreClockMaximumFrequency) && f > frequencyMax) {
        continue;
      }
			if (f > (uint32_t)frequencyReq) {
				error = f - (uint32_t)frequencyReq;
			} else {
				error = (uint32_t)frequencyReq - f;
			}
			if (error < bestError) {
				bestError = error;
				bestFrequency = f;
				bestM = m;
				bestD = d;
			}
		}
	}

	if (bestError == (uint32_t)-1) {
		/* Failed to find a suitable value */
		return ClockWordGeneralFailure;
	}

	*pActualFrequency = bestFrequency;
	*pClockWord = ((bestD - 1) << 16) | (bestM - 1);

	dfDebugPrint(5, ("clockWordLclk: bestFrequency=%lu clockWord=0x%08lx\n", (unsigned long)bestFrequency, (unsigned long)*pClockWord));

	return ClockWordSuccess;
}

static ClockWordStatus
clockWord(
	Adb3CoreDeviceContext* pDevCtx,
	unsigned int clockIndex,
  uint32_t flags,
	uint64_t frequencyReq,
	uint64_t frequencyMin,
	uint64_t frequencyMax,
	uint32_t* pClockWord,
	uint64_t* pActualFrequency)
{
  switch (clockIndex) {
  case 0:
    return clockWordLclk(pDevCtx, flags, pDevCtx->info.referenceClock.admxrc6tl.lclk, frequencyReq, frequencyMin, frequencyMax, pClockWord, pActualFrequency);

  default:
    return ClockWordInvalidIndex;
	}
}

static void
enableInterrupts(
	Adb3CoreDeviceContext* pDevCtx,
	boolean_t bEnable)
{
	Adb3GenericRegs* pRegs;
  DfMemoryHandle hBridge;
  ModelRegsAdmxrc6tx* pModelRegs;
  DfMemoryHandle hModel;

  pRegs = pDevCtx->hardware.bridge.pAdb3;
  hBridge = pDevCtx->hardware.hBridge;
  hModel = pDevCtx->hardware.hModel;
  pModelRegs = pDevCtx->hardware.model.pAdmxrc6tx;

  dfDebugPrint(1, ("enableInterrupts: entered bEnable=%s\n", bEnable ? "TRUE" : "FALSE"));

	if (bEnable) {
    pDevCtx->hwMon.model.admxrc6tl.state = ADMXRC6Tx_HWMONSTATE_ENABLED;
    pDevCtx->hardware.shadow.model.admxrc6tx.intCtl = ADMXRC6Tx_V_INTCTL_FPENABLE | ADMXRC6Tx_V_INTCTL_SYSENABLE;
    dfPciMemWrite32(hModel, &pModelRegs->intCtl, pDevCtx->hardware.shadow.model.admxrc6tx.intCtl);
    dfPciMemRead32(hModel, &pModelRegs->intCtl);
    pDevCtx->hardware.shadow.bridge.adb3.irqEnable = ADB3IRQ_MODEL | ADB3IRQ_DMA_ALL;
    dfPciMemWrite32(hBridge, &pRegs->irqEnable, pDevCtx->hardware.shadow.bridge.adb3.irqEnable);
    dfPciMemRead32(hBridge, &pRegs->irqEnable);
		dfDebugPrint(2, ("enableInterrupts: interrupts enabled, irqEnable=0x%08lx model.intCtl=0x%08lx\n",
      (unsigned long)dfPciMemRead32(hModel, &pRegs->irqEnable), (unsigned long)dfPciMemRead32(hModel, &pModelRegs->intCtl)));
  } else {
    pDevCtx->hwMon.model.admxrc6t1.state = ADMXRC6Tx_HWMONSTATE_DISABLED;
    pDevCtx->hardware.shadow.bridge.adb3.irqEnable = 0;
    dfPciMemWrite32(hModel, &pRegs->irqEnable, pDevCtx->hardware.shadow.bridge.adb3.irqEnable);
    dfPciMemRead32(hModel, &pRegs->irqEnable);
    pDevCtx->hardware.shadow.model.admxrc6tx.intCtl = 0;
    dfPciMemWrite32(hBridge, &pModelRegs->intCtl, pDevCtx->hardware.shadow.model.admxrc6tx.intCtl);
    dfPciMemRead32(hBridge, &pModelRegs->intCtl);
		dfDebugPrint(2, ("enableInterrupts: interrupts disabled, irqEnable=0x%08lx model.intCtl=0x%08lx\n",
      (unsigned long)dfPciMemRead32(hModel, &pRegs->irqEnable), (unsigned long)dfPciMemRead32(hModel, &pModelRegs->intCtl)));
	}
}

static void
flash(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int bankIndex,
  boolean_t bWrite,
  uint64_t location,
  uint32_t* pValue)
{
	ModelRegsAdmxrc6tx* pModelRegs;
  DfMemoryHandle hModel;
  uint32_t flctl, flpage, val32;

  dfDebugPrint(9, ("flash: bankIndex=%lu bWrite=%lu location=0x%08lx_%08lx *pValue=0x%lx\n",
    (unsigned long)bankIndex, (unsigned long)bWrite, dfSplitUint64(location), (unsigned long)*pValue));

  if (bankIndex != 0 || location >= 0x10000000U) {
    return;
  }

  pModelRegs = pDevCtx->hardware.model.pAdmxrc6tx;
  hModel = pDevCtx->hardware.hModel;

	flpage = (uint32_t)location & 0x3fe0000U;
  dfPciMemWrite32(hModel, &pModelRegs->flashAddr, flpage);
	val32 = dfPciMemRead32(hModel, &pModelRegs->flashAddr);

  if (bWrite) {
		/* Select mode 0, set fl_a, Drive fl_a, fl_d, fl_ce_n, fl_oe_n, fl_we_n, assert fl_ce_n */
		flctl = ((uint32_t)location & 0x1fffeU) | (0x7U << 24) | (0x1U << 28);
    dfPciMemWrite32(hModel, &pModelRegs->flashCtl, flctl);
		val32 = dfPciMemRead32(hModel, &pModelRegs->flashCtl);

		/* Set fl_d */
    dfPciMemWrite32(hModel, &pModelRegs->flashData, (uint16_t)*pValue);
		val32 = dfPciMemRead32(hModel, &pModelRegs->flashData);

		/* Assert fl_we_n */
		flctl |= (0x1U << 30);
		dfPciMemWrite32(hModel, &pModelRegs->flashCtl, flctl);
		val32 = dfPciMemRead32(hModel, &pModelRegs->flashCtl);

		/* Deassert fl_we_n */
		flctl &= ~(0x1U << 30);
		dfPciMemWrite32(hModel, &pModelRegs->flashCtl, flctl);
		val32 = dfPciMemRead32(hModel, &pModelRegs->flashCtl);

		/* Deassert fl_ce_n */
		flctl &= ~(0x1U << 28);
		dfPciMemWrite32(hModel, &pModelRegs->flashCtl, flctl);
		val32 = dfPciMemRead32(hModel, &pModelRegs->flashCtl);

		/* Stop driving fl_a, fl_d, fl_ce_n, fl_oe_n, fl_we_n */
		flctl &= ~(0x7U << 24);
		dfPciMemWrite32(hModel, &pModelRegs->flashCtl, flctl);
		val32 = dfPciMemRead32(hModel, &pModelRegs->flashCtl);
	} else {
		/* Select mode 0, set fl_a, Drive fl_a, fl_ce_n, fl_oe_n, fl_we_n, assert fl_ce_n */
		flctl = ((uint32_t)location & 0x1fffeU) | (0x3U << 25) | (0x1U << 28);
		dfPciMemWrite32(hModel, &pModelRegs->flashCtl, flctl);
		val32 = dfPciMemRead32(hModel, &pModelRegs->flashCtl);

		/* Assert fl_oe_n */
		flctl |= (0x1U << 29);
		dfPciMemWrite32(hModel, &pModelRegs->flashCtl, flctl);
		val32 = dfPciMemRead32(hModel, &pModelRegs->flashCtl);

		/* Read fl_d */
    *pValue = (uint16_t)dfPciMemRead32(hModel, &pModelRegs->flashData);

		/* Deassert fl_oe_n */
		flctl &= ~(0x1U << 29);
		dfPciMemWrite32(hModel, &pModelRegs->flashCtl, flctl);
		val32 = dfPciMemRead32(hModel, &pModelRegs->flashCtl);

		/* Deassert fl_ce_n */
		flctl &= ~(0x1U << 28);
		dfPciMemWrite32(hModel, &pModelRegs->flashCtl, flctl);
		val32 = dfPciMemRead32(hModel, &pModelRegs->flashCtl);

		/* Stop driving fl_a, fl_ce_n, fl_oe_n, fl_we_n */
		flctl &= ~(0x3U << 25);
		dfPciMemWrite32(hModel, &pModelRegs->flashCtl, flctl);
		val32 = dfPciMemRead32(hModel, &pModelRegs->flashCtl);

		dfDebugPrint(9, ("flash: *value=0x%lx\n", (unsigned long)*pValue));
	}
}


static void
flashReset(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int bankIndex)
{
	uint32_t flctl;
	ModelRegsAdmxrc6tx* pModelRegs;
  DfMemoryHandle hModel;

	dfDebugPrint(9, ("flashReset: entered\n"));

  if (bankIndex != 0) {
    return;
  }

  pModelRegs = pDevCtx->hardware.model.pAdmxrc6tx;
  hModel = pDevCtx->hardware.hModel;

	/* Drive fl_rst_n, assert fl_rst_n */
	flctl = (0x1U << 27) | (0x1U << 31);
  dfPciMemWrite32(hModel, &pModelRegs->flashCtl, flctl);
	dfPciMemRead32(hModel, &pModelRegs->flashCtl);

	/* Deassert fl_rst_n */
	flctl &= ~(0x1U << 31);
	dfPciMemWrite32(hModel, &pModelRegs->flashCtl, flctl);
	dfPciMemRead32(hModel, &pModelRegs->flashCtl);

	/* Stop driving fl_rst_n */
	flctl &= ~(0x1U << 27);
	dfPciMemWrite32(hModel, &pModelRegs->flashCtl, flctl);
	dfPciMemRead32(hModel, &pModelRegs->flashCtl);

  /* Allow device time to recover */
  dfDelayThreadFor(dfMicrosecondsToTime(20));
}

static FpgaControlStatus
fpgaControl(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int index,
  CoreFpgaControlOp opCode,
  boolean_t* pbValue)
{
  ModelRegsAdmxrc6tx* pModelRegs;
  DfMemoryHandle hModel;
  uint32_t val32;
  boolean_t bOk;

  dfDebugPrint(8, ("fpgaControl: index=%lu opCode=%lu pbValue=0x%p\n",
    (unsigned long)index, (unsigned long)opCode, pbValue));
  
  if (0 != index) {
    return FpgaControlInvalidIndex;
  }

  pModelRegs = pDevCtx->hardware.model.pAdmxrc6tx;
  hModel = pDevCtx->hardware.hModel;

  switch (opCode) {
  case CoreFpgaControlOpGetDone:
    bOk = TRUE;
    val32 = dfPciMemRead32(hModel, &pModelRegs->fpgaCtl);
    if (!(val32 & 0x4U)) {
      dfDebugPrint(9, ("fpgaControl: fpgaStat=0x%lx\n", (unsigned long)val32));
      bOk = FALSE;
    }
    *pbValue = bOk;
    dfDebugPrint(9, ("fpgaControl: opCode=GetDone val32=0x%lx *pbValue=%s\n", (unsigned long)val32, bOk ? "TRUE" : "FALSE"));
    break;

  case CoreFpgaControlOpGetLinkStatus:
    bOk = TRUE;
    /* Check that MPTL is up */
    val32 = dfPciMemRead32(hModel, &pModelRegs->mptlStat);
    if ((val32 & 0x7U) != 0x7U) {
      dfDebugPrint(9, ("fpgaControl: mptlStat=0x%lx\n", (unsigned long)val32));
      bOk = FALSE;
    }
    *pbValue = bOk;
    dfDebugPrint(9, ("fpgaControl: opCode=GetLinkStat val32=0x%lx *pbValue=%s\n", (unsigned long)val32, bOk ? "TRUE" : "FALSE"));
    break;

  case CoreFpgaControlOpGetInit:
    val32 = dfPciMemRead32(hModel, &pModelRegs->fpgaCtl);
    *pbValue = (val32 & 0x2U) ? TRUE : FALSE;
    dfDebugPrint(9, ("fpgaControl: opCode=GetInit val32=0x%lx *pbValue=%s\n", (unsigned long)val32, *pbValue ? "TRUE" : "FALSE"));
    break;

  case CoreFpgaControlOpSetProg:
    val32 = pDevCtx->hardware.shadow.model.admxrc6tx.fpgaCtl;
    if (*pbValue) {
      val32 |= 0x1U;
    } else {
      val32 &= ~0x1U;
    }
    dfDebugPrint(9, ("fpgaControl: opCode=SetProg val32=0x%lx *pbValue=%s\n", (unsigned long)val32, *pbValue ? "TRUE" : "FALSE"));
    dfPciMemWrite32(hModel, &pModelRegs->fpgaCtl, val32);
    dfPciMemRead32(hModel, &pModelRegs->fpgaCtl);
    break;

  default:
    return FpgaControlInvalidOp;
  }

  return FpgaControlSuccess;
}

static FpgaSelectMapStatus
fpgaSelectMap(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int index,
  boolean_t bWrite,
  size_t length,
  void* pBuffer)
{
  const size_t mask32 = 0x3U;
  ModelRegsAdmxrc6tx* pModelRegs;
  DfMemoryHandle hModel;
  uint8_t* pBuffer8;
  uint32_t* pBuffer32;
  size_t length32, align32;
  uint8_t val8;
  uint32_t val32;

  dfDebugPrint(8, ("fpgaSelectMap: index=%lu bWrite=%s length=%llu(0x%08lx_%08lx) pBuffer=0x%p\n",
    (unsigned long)index, bWrite ? "TRUE" : "FALSE", (unsigned long long)length, dfSplitUint64((uint64_t)length), pBuffer));
  
  if (0 != index) {
    return FpgaSelectMapInvalidIndex;
  }

  pModelRegs = pDevCtx->hardware.model.pAdmxrc6tx;
  hModel = pDevCtx->hardware.hModel;
  pBuffer8 = (uint8_t*)pBuffer;

	if (bWrite) {
    /* Align pointer to 32-bit boundary using 8-bit transfers */
    align32 = (size_t)(uintptr_t)pBuffer & mask32;
    length -= align32;
    while (align32--) {
			val8 = *pBuffer8++;
			dfPciMemWrite8(hModel, &pModelRegs->selectMap[0], val8);
    }

    /* Do bulk of data using 32-bit transfers */
    length32 = length & ~mask32;
    pBuffer32 = (uint32_t*)pBuffer8;
    length -= length32;
    while (length32) {
      val32 = dfLe32ToCpu(*pBuffer32++);
			dfPciMemWrite32(hModel, (uint32_t*)&pModelRegs->selectMap[0], val32);
      length32 -= 4;
    }

    /* Do remainder using 8-bit transfers */
    pBuffer8 = (uint8_t*)pBuffer32;
		while (length--) {
			val8 = *pBuffer8++;
			dfPciMemWrite8(hModel, &pModelRegs->selectMap[0], val8);
		}

		dfPciMemRead32(hModel, &pModelRegs->fpgaCtl); /* Ensure above writes have propagated to destination */
	} else {
		while (length--) {
			val8 = dfPciMemRead8(hModel, &pModelRegs->selectMap[0]);
			*pBuffer8++ = val8;
		}
	}

  return FpgaSelectMapSuccess;
}

static boolean_t
initHardware(
	Adb3CoreDeviceContext* pDevCtx,
	boolean_t bInit)
{
  Adb3GenericRegs* pRegs;
  DfMemoryHandle hBridge;
  ModelRegsAdmxrc6tx* pModelRegs;
  DfMemoryHandle hModel;
  unsigned int i;

  dfDebugPrint(2, ("initHardware: bInit=%s\n", bInit ? "TRUE" : "FALSE"));

  pRegs = pDevCtx->hardware.bridge.pAdb3;
  hBridge = pDevCtx->hardware.hBridge;
  pModelRegs = pDevCtx->hardware.model.pAdmxrc6tx;
  hModel = pDevCtx->hardware.hModel;

  if (bInit) {
    pDevCtx->info.model = CoreModelAdmxrc6tl;

    /* Initialize DMA enables */
    dfPciMemWrite32(hBridge, &pRegs->dmaEnable, ADB3DMAEN_LOCAL_MASTER | ADB3DMAEN_DMA_ALL);

    /* Allow up to 3 PCIe packets in flight at a time */
    dfPciMemWrite32(hBridge, &pRegs->pcieControl, 0x02100100U);           

    /* Initialize DMA engine registers */
    for (i = 0; i < pDevCtx->info.bootstrap.numDmaChannel; i++) {
      /* Request up to 32 x 16-byte words on PCIe side */
      dfPciMemWrite32(hBridge, &pRegs->dmaEngine[i].hostControl, 0x20U);    
      /* Allow up to 64 bytes read and 64 bytes written on MPTL */
      dfPciMemWrite32(hBridge, &pRegs->dmaEngine[i].fpgaControl, 0x80008U);
    }

    /* Initialize FPGACTL register for 8-bit SelectMap port */
    pDevCtx->hardware.shadow.model.admxrc6tx.fpgaCtl = 0;
    dfPciMemWrite32(hModel, &pModelRegs->fpgaCtl, pDevCtx->hardware.shadow.model.admxrc6tx.fpgaCtl);
		dfPciMemRead32(hModel, &pModelRegs->fpgaCtl);

    /* Identify Flash device */
    if (pDevCtx->info.bootstrap.flashWidth != 0) {
      if (flashCfiIdentify(pDevCtx, 0, 0)) {
        pDevCtx->info.flash.bPresent = TRUE;
        pDevCtx->info.flash.targetArea.start = 0x1200000U;
        pDevCtx->info.flash.targetArea.length = pDevCtx->info.flash.totalSize - pDevCtx->info.flash.targetArea.start;
        readVpd(pDevCtx);
      } else {
        dfDebugPrint(0, ("*** initHardware: failed to identify Flash\n"));
      }
    }

    /* Identify any XRM module fitted */
    admxrc6txIdentifyXrm(pDevCtx);

    /* Divide LCLK reference clock by 2 because on the ADM-XRC-6TL, a CLKDIV value of 2 is used in the PLLs that generate LCLK */
    pDevCtx->info.referenceClock.admxrc6tl.lclk = pDevCtx->info.vpd.data.admxrc6tl.rev1.lclkRef / 2;

    /* Find base addresses and sizes of local bus windows */
    pDevCtx->info.window[0].busBase = pDevCtx->rawResources.variant.pci.memBar[2].base;
    pDevCtx->info.window[0].busSize = pDevCtx->rawResources.variant.pci.memBar[2].size;
    pDevCtx->info.window[0].localBase = 0;
    pDevCtx->info.window[0].localSize = pDevCtx->translatedResources.variant.pci.memBar[2].size;
    pDevCtx->info.window[0].translatedBase = pDevCtx->translatedResources.variant.pci.memBar[2].base;
    pDevCtx->info.window[0].translatedSize = pDevCtx->translatedResources.variant.pci.memBar[2].size;
    pDevCtx->info.window[0].pKernelBase = pDevCtx->hardware.pBridged;
    pDevCtx->info.window[0].kernelSize = (size_t)pDevCtx->translatedResources.variant.pci.memBar[2].size;
    pDevCtx->info.window[1].busBase = pDevCtx->rawResources.variant.pci.memBar[3].base;
    pDevCtx->info.window[1].busSize = pDevCtx->rawResources.variant.pci.memBar[3].size;
    pDevCtx->info.window[1].localBase = 0;
    pDevCtx->info.window[1].localSize = pDevCtx->translatedResources.variant.pci.memBar[3].size;
    pDevCtx->info.window[1].translatedBase = pDevCtx->translatedResources.variant.pci.memBar[3].base;
    pDevCtx->info.window[1].translatedSize = pDevCtx->translatedResources.variant.pci.memBar[3].size;
    pDevCtx->info.window[1].pKernelBase = pDevCtx->hardware.pBridgedPrefetch;
    pDevCtx->info.window[1].kernelSize = (size_t)pDevCtx->translatedResources.variant.pci.memBar[3].size;
    pDevCtx->info.window[2].busBase = pDevCtx->rawResources.variant.pci.memBar[1].base;
    pDevCtx->info.window[2].busSize = pDevCtx->rawResources.variant.pci.memBar[1].size;
    pDevCtx->info.window[2].localBase = 0;
    pDevCtx->info.window[2].localSize = 0;
    pDevCtx->info.window[2].translatedBase = pDevCtx->translatedResources.variant.pci.memBar[1].base;
    pDevCtx->info.window[2].translatedSize = pDevCtx->translatedResources.variant.pci.memBar[1].size;
    pDevCtx->info.window[2].pKernelBase = pDevCtx->hardware.model.pGeneric;
    pDevCtx->info.window[2].kernelSize = (size_t)pDevCtx->translatedResources.variant.pci.memBar[1].size;
    pDevCtx->info.window[3].busBase = pDevCtx->rawResources.variant.pci.memBar[0].base;
    pDevCtx->info.window[3].busSize = pDevCtx->rawResources.variant.pci.memBar[0].size;
    pDevCtx->info.window[3].localBase = 0;
    pDevCtx->info.window[3].localSize = 0;
    pDevCtx->info.window[3].translatedBase = pDevCtx->translatedResources.variant.pci.memBar[0].base;
    pDevCtx->info.window[3].translatedSize = pDevCtx->translatedResources.variant.pci.memBar[0].size;
    pDevCtx->info.window[3].pKernelBase = pDevCtx->hardware.bridge.pGeneric;
    pDevCtx->info.window[3].kernelSize = (size_t)pDevCtx->translatedResources.variant.pci.memBar[0].size;

    /* Initialize sensor information structures */
    admxrc6txInitSensorInfo(pDevCtx);
  } else {
    /* Deinitialize DMA enables */
    dfPciMemWrite32(hBridge, &pRegs->dmaEnable, 0);
  }

  return TRUE;
}

static boolean_t
isr(
  DfInterruptObject* pInterruptObj,
  void* pInterruptContext)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pInterruptContext;
  Adb3GenericRegs* pRegs;
  DfMemoryHandle hBridge;
  ModelRegsAdmxrc6tx* pModelRegs;
  DfMemoryHandle hModel;
  uint32_t irqStatus, dmaIrqStatus;
  uint32_t intStat;
  unsigned int channel, n;

  dfDebugPrint(8, ("isr: entered\n"));

  pRegs = pDevCtx->hardware.bridge.pAdb3;
  hBridge = pDevCtx->hardware.hBridge;
  pModelRegs = pDevCtx->hardware.model.pAdmxrc6tx;
  hModel = pDevCtx->hardware.hModel;

  /* Sample the currently active interrupts, and mask the value with the interrupts we want */
  irqStatus = dfPciMemRead32(hBridge, &pRegs->irqStatus) & pDevCtx->hardware.shadow.bridge.adb3.irqEnable;

  /* Clear the interrupts we saw */
  dfPciMemWrite32(hBridge, &pRegs->irqStatus, irqStatus);

  if (irqStatus & ADB3IRQ_MODEL) {
    dfDebugPrint(9, ("isr: model-specific interrupt\n"));
    intStat = dfPciMemRead32(hModel, &pModelRegs->intStat) & pDevCtx->hardware.shadow.model.admxrc6tx.intCtl & ADMXRC6Tx_M_INTCTL_ENABLE;
    /* Clear the model-specific interrupts that we saw */
    dfPciMemWrite32(hModel, &pModelRegs->intStat, intStat);
    if (intStat & ADMXRC6Tx_V_INTSTAT_FP) {
      dfDebugPrint(9, ("isr: target FPGA 0 interrupt\n"));
      dfDpcSchedule(&pDevCtx->interrupt.fpga[0].dpc, (void*)(uintptr_t)0);
    }
    if (intStat & ADMXRC6Tx_V_INTSTAT_SYS) {
      dfAssert(pDevCtx->hwMon.model.admxrc6tl.state == ADMXRC6Tx_HWMONSTATE_ENABLED);
      dfDebugPrint(9, ("isr: hardware monitor interrupt\n"));
      dfDpcSchedule(&pDevCtx->interrupt.hwMon[0].dpc, (void*)(uintptr_t)0);
      /* Disable hardware monitor interrupt */
      pDevCtx->hardware.shadow.model.admxrc6tx.intCtl &= ~ADMXRC6Tx_V_INTCTL_SYSENABLE;
      dfPciMemWrite32(hModel, &pModelRegs->intCtl, pDevCtx->hardware.shadow.model.admxrc6tx.intCtl);
      dfPciMemRead32(hModel, &pModelRegs->intCtl);
      pDevCtx->hwMon.model.admxrc6tl.state = ADMXRC6Tx_HWMONSTATE_DISABLED;
    }
  }

  n = pDevCtx->info.bootstrap.numDmaChannel;
  for (channel = 0; channel < n; channel++) {
    if (irqStatus & (ADB3IRQ_DMA0 << channel)) {
      dmaIrqStatus = dfPciMemRead32(hBridge, &pRegs->dmaEngine[channel].irq.status);
      dfDebugPrint(9, ("isr: DMA%lu interrupt, DMA%lu.irqStatus=0x%08lx\n",
        (unsigned long)channel, (unsigned long)channel, (unsigned long)dmaIrqStatus));
      dfPciMemWrite32(hBridge, &pRegs->dmaEngine[channel].irq.ack, dmaIrqStatus);
      dfDpcSchedule(&pDevCtx->dma.channel[channel].completionDpc, (void*)(uintptr_t)channel);
    }
  }

  return irqStatus ? TRUE : FALSE;
}

static boolean_t
mapBusResources(
	Adb3CoreDeviceContext* pDevCtx,
	DfBusResources* pRaw,
	DfBusResources* pTranslated,
	boolean_t bUnmap)
{
	dfDebugPrint(1, ("mapBusResources: bUnmap=%s\n", bUnmap ? "TRUE" : "FALSE"));

	if (bUnmap) {
		if (NULL != pDevCtx->hardware.bridge.pGeneric) {
			if (!dfIoSpaceUnmap(pDevCtx->pDevObj, pDevCtx->hardware.bridge.pGeneric)) {
				dfDebugPrint(0, ("*** mapBusResources: failed to unmap ADB3 generic registers\n"));
			}
      pDevCtx->hardware.bridge.pGeneric = NULL;
		}
		if (NULL != pDevCtx->hardware.model.pGeneric) {
			if (!dfIoSpaceUnmap(pDevCtx->pDevObj, pDevCtx->hardware.model.pGeneric)) {
				dfDebugPrint(0, ("*** mapBusResources: failed to unmap ADB3 model-specific registers\n"));
			}
      pDevCtx->hardware.model.pGeneric = NULL;
		}
    if (NULL != pDevCtx->hardware.pBridged) {
			if (!dfIoSpaceUnmap(pDevCtx->pDevObj, pDevCtx->hardware.pBridged)) {
				dfDebugPrint(0, ("*** mapBusResources: failed to unmap ADB3 bridged space\n"));
			}
      pDevCtx->hardware.pBridged = NULL;
		}
    if (NULL != pDevCtx->hardware.pBridgedPrefetch) {
			if (!dfIoSpaceUnmap(pDevCtx->pDevObj, pDevCtx->hardware.pBridgedPrefetch)) {
				dfDebugPrint(0, ("*** mapBusResources: failed to unmap ADB3 prefetchable bridged space\n"));
			}
      pDevCtx->hardware.pBridgedPrefetch = NULL;
		}
		dfDebugPrint(2, ("mapBusResources: unmapped\n"));
	} else {
		pDevCtx->hardware.bridge.pGeneric = dfIoSpaceMap(pDevCtx->pDevObj, pTranslated, 0, 0, pTranslated->variant.pci.memBar[0].size, &pDevCtx->hardware.hBridge);
		if (NULL == pDevCtx->hardware.bridge.pGeneric) {
			dfDebugPrint(0, ("*** mapBusResources: failed to map ADB3 generic registers\n"));
			return FALSE;
		} else {
			dfDebugPrint(2, ("mapBusResources: mapped ADB3 generic registers @ 0x%08lx_%08lx to 0x%p\n",
        dfSplitUint64(pTranslated->variant.pci.memBar[0].base), pDevCtx->hardware.bridge.pGeneric));
		}
		pDevCtx->hardware.model.pGeneric = dfIoSpaceMap(pDevCtx->pDevObj, pTranslated, 1, 0, pTranslated->variant.pci.memBar[1].size, &pDevCtx->hardware.hModel);
		if (NULL == pDevCtx->hardware.model.pGeneric) {
			dfDebugPrint(0, ("*** mapBusResources: failed to map ADB3 model-specific registers\n"));
			return FALSE;
		} else {
			dfDebugPrint(2, ("mapBusResources: mapped ADB3 model-specific registers @ 0x%08lx_%08lx to 0x%p\n",
        dfSplitUint64(pTranslated->variant.pci.memBar[1].base), pDevCtx->hardware.model.pGeneric));
		}
    pDevCtx->hardware.pBridged = dfIoSpaceMap(pDevCtx->pDevObj, pTranslated, 2, 0, pTranslated->variant.pci.memBar[2].size, &pDevCtx->hardware.hBridged);
		if (NULL == pDevCtx->hardware.pBridged) {
			dfDebugPrint(0, ("*** mapBusResources: failed to map ADB3 bridged space\n"));
			return FALSE;
		} else {
			dfDebugPrint(2, ("mapBusResources: mapped ADB3 bridged space @ 0x%08lx_%08lx to 0x%p\n",
        dfSplitUint64(pTranslated->variant.pci.memBar[2].base), pDevCtx->hardware.pBridged));
		}
    pDevCtx->hardware.pBridgedPrefetch = dfIoSpaceMap(pDevCtx->pDevObj, pTranslated, 3, 0, pTranslated->variant.pci.memBar[3].size, &pDevCtx->hardware.hBridgedPrefetch);
		if (NULL == pDevCtx->hardware.pBridgedPrefetch) {
			dfDebugPrint(0, ("*** mapBusResources: failed to map ADB3 prefetchable bridged space\n"));
			return FALSE;
		} else {
			dfDebugPrint(2, ("mapBusResources: mapped ADB3 prefetchable bridged space @ 0x%08lx_%08lx to 0x%p\n",
        dfSplitUint64(pTranslated->variant.pci.memBar[3].base), pDevCtx->hardware.pBridgedPrefetch));
		}
		dfDebugPrint(2, ("mapBusResources: mapped OK\n"));
	}

	return TRUE;
}

static CoreVpdStatus
readWriteVpd(
	Adb3CoreDeviceContext* pDevCtx,
	boolean_t bWrite,
	size_t address,
	size_t length,
	void* pData)
{
  CoreFlashStatus flashStatus;

  if (CHECK_BOUNDS(address, length, pDevCtx->info.flash.totalSize)) {
    return CoreVpdInvalidRegion;
  }
	if (bWrite) {
    flashStatus = flashWrite(pDevCtx, 0, address, length, pData);
	} else {
		flashStatus = flashRead(pDevCtx, 0, address, length, pData);
	}
  if (CoreFlashInvalidIndex == flashStatus) {
    /* Should never happen, but map CoreFlashInvalidIndex to CoreVpdGeneralFailure */
    return CoreVpdGeneralFailure;
  } else {
    return (CoreVpdStatus)flashStatus;
  }
}

static boolean_t
validateBusResources(
	Adb3CoreDeviceContext* pDevCtx,
	DfBusResources* pRaw,
	DfBusResources* pTranslated)
{
	dfDebugPrint(1, ("validateBusResources: entered\n"));

	if (pTranslated->variant.pci.numMemBar < 4) {
		dfDebugPrint(0, ("*** validateBusResources: less than 4 memory BARs\n"));
		return FALSE;
	}
	if (pTranslated->variant.pci.memBar[0].size < 0x1000U) {
		dfDebugPrint(0, ("*** validateBusResources: BAR0 smaller than 0x1000 bytes\n"));
		return FALSE;
	}
	if (pTranslated->variant.pci.memBar[1].size < sizeof(ModelRegsAdmxrc6tx)) {
		dfDebugPrint(0, ("*** validateBusResources: BAR1 smaller than 0x%lu bytes\n", (unsigned long)sizeof(ModelRegsAdmxrc6tx)));
		return FALSE;
	}
	if (pTranslated->variant.pci.memBar[2].size < 0x100000U) {
		dfDebugPrint(0, ("*** validateBusResources: BAR2 smaller than 0x100000 bytes\n"));
		return FALSE;
	}
	if (pTranslated->variant.pci.memBar[3].size < 0x100000U) {
		dfDebugPrint(0, ("*** validateBusResources: BAR3 smaller than 0x100000 bytes\n"));
		return FALSE;
	}
	if (pTranslated->variant.pci.numIrq != 1) {
		dfDebugPrint(0, ("*** validateBusResources: not exactly 1 interrupt\n"));
		return FALSE;
	}

	dfDebugPrint(2, ("validateBusResources: OK\n"));

	return TRUE;
}

void
adb3CoreBootstrapAdmxrc6tl(
	Adb3CoreDeviceContext* pDevCtx)
{
  dfDebugPrint(1, ("adb3CoreBootstrapAdmxrc6tl: entered\n"));

  dfAssert(DF_OFFSET_OF(VpdAdmxrc6tlRev0, checksum) == sizeof(VpdAdmxrc6tlRev0) - 1);

  pDevCtx->methods.pClockProgram = clockProgram;
  pDevCtx->methods.pClockWord = clockWord;
  pDevCtx->methods.pDmaList = adb3DmaList;
  pDevCtx->methods.pDmaTransfer = adb3DmaTransfer;
  pDevCtx->methods.pEnableInterrupts = enableInterrupts;
  pDevCtx->methods.pFlash = flash;
  pDevCtx->methods.pFlashReset = flashReset;
  pDevCtx->methods.pFpgaControl = fpgaControl;
  pDevCtx->methods.pFpgaSelectMap = fpgaSelectMap;
  pDevCtx->methods.pInitHardware = initHardware;
  pDevCtx->methods.pIsr = isr;
  pDevCtx->methods.pMapBusResources = mapBusResources;
  pDevCtx->methods.pReadSensor = admxrc6txReadSensor;
  pDevCtx->methods.pReadWriteVpd = readWriteVpd;
  pDevCtx->methods.pValidateBusResources = validateBusResources;

  pDevCtx->info.bootstrap.numClockGen = 1;
  pDevCtx->info.bootstrap.numDmaChannel = (pDevCtx->pDevObj->hardwareId.variant.pci.subDevice == PCI_SUBDEVICE_ADB3_ADMXRC6TL_FX50T) ? 2 : 1;
  pDevCtx->info.bootstrap.numSensor = ADMXRC6TX_NUM_SENSOR;
  dfParameterGetBoolean(pDevCtx->pDevObj->pDrvObj, "IncludeDiagSensors", &pDevCtx->info.bootstrap.bIncludeDiagSensors);
  if (pDevCtx->info.bootstrap.bIncludeDiagSensors) {
    pDevCtx->info.bootstrap.numSensor += g_admxrc6txNumDiagSensor;
  }
  pDevCtx->info.bootstrap.numIoModule = 1;
  pDevCtx->info.bootstrap.numTargetFpga = 1;
  pDevCtx->info.bootstrap.numWindow = 4;
  pDevCtx->info.bootstrap.bVpdInFlash = TRUE;
  pDevCtx->info.bootstrap.vpdLength = 0x200000U;
  pDevCtx->info.bootstrap.vpdOffset = 0x1000000U;
  pDevCtx->info.bootstrap.flashDeviceWidth = 2;
  pDevCtx->info.bootstrap.flashWidth = 2;
  pDevCtx->info.bootstrap.bFlashStackedDie = FALSE;
  pDevCtx->info.bootstrap.b64BitDmaAddress = FALSE;
  dfParameterGetBoolean(pDevCtx->pDevObj->pDrvObj, "PciAddress64Bit", &pDevCtx->info.bootstrap.b64BitDmaAddress);
  pDevCtx->info.bootstrap.dmaDescriptorSize = 32;
  pDevCtx->info.bootstrap.dmaTransferMaxSize = 0x1000000U;
  pDevCtx->info.bootstrap.bWorkAround4kCrossing = (pDevCtx->pDevObj->hardwareId.variant.pci.revision <= 0x01U) ? TRUE : FALSE;
}
